import spacy
import bert_score
import numpy as np
import torch
from typing import Dict, List, Set, Tuple, Union
from transformers import logging
logging.set_verbosity_error()

from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
from transformers import LongformerTokenizer, LongformerForMultipleChoice, LongformerForSequenceClassification
from transformers import DebertaV2ForSequenceClassification, DebertaV2Tokenizer
from selfcheckgpt.utils import MQAGConfig, expand_list1, expand_list2, NLIConfig
from selfcheckgpt.modeling_mqag import question_generation_sentence_level, answering
from selfcheckgpt.modeling_ngram import UnigramModel, NgramModel

# ---------------------------------------------------------------------------------------- #
# Functions for counting
def method_simple_counting(
    prob,
    u_score,
    prob_s,
    u_score_s,
    num_samples,
    AT,
):
    """
    simple counting method score => count_mismatch / (count_match + count_mismatch)
    :return score: 'inconsistency' score
    """
    # bad questions, i.e. not answerable given the passage
    if u_score < AT:
        return 0.5
    a_DT = np.argmax(prob)
    count_good_sample, count_match = 0, 0
    for s in range(num_samples):
        if u_score_s[s] >= AT:
            count_good_sample += 1
            a_S = np.argmax(prob_s[s])
            if a_DT == a_S:
                count_match += 1
    if count_good_sample == 0:
        score = 0.5
    else:
        score = (count_good_sample-count_match) / count_good_sample
    return score

def method_vanilla_bayes(
    prob,
    u_score,
    prob_s,
    u_score_s,
    num_samples,
    beta1, beta2, AT,
):
    """
    (vanilla) bayes method score: compute P(sentence is non-factual | count_match, count_mismatch)
    :return score: 'inconsistency' score
    """
    if u_score < AT:
        return 0.5
    a_DT = np.argmax(prob)
    count_match, count_mismatch = 0, 0
    for s in range(num_samples):
        if u_score_s[s] >= AT:
            a_S = np.argmax(prob_s[s])
            if a_DT == a_S:
                count_match += 1
            else:
                count_mismatch += 1
    gamma1 = beta2 / (1.0-beta1)
    gamma2 = beta1 / (1.0-beta2)
    score = (gamma2**count_mismatch) / ((gamma1**count_match) + (gamma2**count_mismatch))
    return score

def method_bayes_with_alpha(
    prob,
    u_score,
    prob_s,
    u_score_s,
    num_samples,
    beta1, beta2,
):
    """
    bayes method (with answerability score, i.e. soft-counting) score
    :return score: 'inconsistency' score
    """
    a_DT = np.argmax(prob)
    count_match, count_mismatch = 0, 0
    for s in range(num_samples):
        ans_score = u_score_s[s]
        a_S = np.argmax(prob_s[s])
        if a_DT == a_S:
            count_match += ans_score
        else:
            count_mismatch += ans_score
    gamma1 = beta2 / (1.0-beta1)
    gamma2 = beta1 / (1.0-beta2)
    score = (gamma2**count_mismatch) / ((gamma1**count_match) + (gamma2**count_mismatch))
    return score

def answerability_scoring(
    u_model,
    u_tokenizer,
    question,
    context,
    max_length,
    device,
):
    """
    :return prob: prob -> 0.0 means unanswerable, prob -> 1.0 means answerable
    """
    input_text = question + ' ' + u_tokenizer.sep_token + ' ' + context
    inputs = u_tokenizer(input_text, max_length=max_length, truncation=True, return_tensors="pt")
    inputs = inputs.to(device)
    logits = u_model(**inputs).logits
    logits = logits.squeeze(-1)
    prob = torch.sigmoid(logits).item()
    return prob

class SelfCheckMQAG:
    """
    SelfCheckGPT (MQAG varaint): Checking LLM's text against its own sampled texts via MultipleChoice Question Answering
    """
    def __init__(
        self,
        g1_model: str = None,
        g2_model: str = None,
        answering_model: str = None,
        answerability_model: str = None,
        device = None
    ):

        g1_model = g1_model if g1_model is not None else MQAGConfig.generation1_squad
        g2_model = g2_model if g2_model is not None else MQAGConfig.generation2
        answering_model = answering_model if answering_model is not None else MQAGConfig.answering
        answerability_model = answerability_model if answerability_model is not None else MQAGConfig.answerability

        # Question Generation Systems (G1 & G2)
        self.g1_tokenizer = AutoTokenizer.from_pretrained(g1_model)
        self.g1_model = AutoModelForSeq2SeqLM.from_pretrained(g1_model)
        self.g2_tokenizer = AutoTokenizer.from_pretrained(g2_model)
        self.g2_model = AutoModelForSeq2SeqLM.from_pretrained(g2_model)

        # Question Answering System (A)
        self.a_tokenizer = LongformerTokenizer.from_pretrained(answering_model)
        self.a_model = LongformerForMultipleChoice.from_pretrained(answering_model)

        # (Un)Answerability System (U)
        self.u_tokenizer = LongformerTokenizer.from_pretrained(answerability_model)
        self.u_model = LongformerForSequenceClassification.from_pretrained(answerability_model)

        self.g1_model.eval()
        self.g2_model.eval()
        self.a_model.eval()
        self.u_model.eval()

        if device is None:
            device = torch.device("cpu")
        self.g1_model.to(device)
        self.g2_model.to(device)
        self.a_model.to(device)
        self.u_model.to(device)
        self.device = device
        print("SelfCheck-MQAG initialized to device", device)

    @torch.no_grad()
    def predict(
        self,
        sentences: List[str],
        passage: str,
        sampled_passages: List[str],
        num_questions_per_sent: int = 5,
        scoring_method: str = "bayes_with_alpha",
        **kwargs,
    ):
        """
        This function takes sentences (to be evaluated) with sampled passages (evidence), and return sent-level scores
        :param sentences: list[str] -- sentences to be evaluated, e.g. GPT text response spilt by spacy
        :param passage: str -- the passage to be evaluated, note that splitting(passage) ---> sentences
        :param sampled_passages: list[str] -- stochastically generated responses (without sentence splitting)
        :param num_questions_per_sent: int -- number of quetions to be generated per sentence
        :return sent_scores: sentence-level score of the same length as len(sentences) # inconsistency_score, i.e. higher means likely hallucination
        """
        assert scoring_method in ['counting', 'bayes', 'bayes_with_alpha']
        num_samples = len(sampled_passages)
        sent_scores = []
        for sentence in sentences:

            # Question + Choices Generation
            questions = question_generation_sentence_level(
                self.g1_model, self.g1_tokenizer,
                self.g2_model, self.g2_tokenizer,
                sentence, passage, num_questions_per_sent, self.device)

            # Answering
            scores = []
            max_seq_length = 4096 # answering & answerability max length
            for question_item in questions:
                question, options = question_item['question'], question_item['options']
                # response
                prob = answering(
                    self.a_model, self.a_tokenizer,
                    question, options, passage,
                    max_seq_length, self.device)

                u_score = answerability_scoring(
                    self.u_model, self.u_tokenizer,
                    question, passage,
                    max_seq_length, self.device)

                prob_s = np.zeros((num_samples, 4))
                u_score_s = np.zeros((num_samples,))
                for si, sampled_passage in enumerate(sampled_passages):

                    # sample
                    prob_s[si] = answering(
                        self.a_model, self.a_tokenizer,
                        question, options, sampled_passage,
                        max_seq_length, self.device)
                    u_score_s[si] = answerability_scoring(
                        self.u_model, self.u_tokenizer,
                        question, sampled_passage,
                        max_seq_length, self.device)

                # doing comparision
                if scoring_method == 'counting':
                    score = method_simple_counting(prob, u_score, prob_s, u_score_s, num_samples, AT=kwargs['AT'])
                elif scoring_method == 'bayes':
                    score = method_vanilla_bayes(prob, u_score, prob_s, u_score_s, num_samples, beta1=kwargs['beta1'], beta2=kwargs['beta2'], AT=kwargs['AT'])
                elif scoring_method == 'bayes_with_alpha':
                    score = method_bayes_with_alpha(prob, u_score, prob_s, u_score_s, num_samples, beta1=kwargs['beta1'], beta2=kwargs['beta2'])
                scores.append(score)
            sent_score = np.mean(scores)
            sent_scores.append(sent_score)

        return np.array(sent_scores)

class SelfCheckBERTScore:
    """
    SelfCheckGPT (BERTScore variant): Checking LLM's text against its own sampled texts via BERTScore (against best-matched sampled sentence)
    """
    def __init__(self, default_model="en", rescale_with_baseline=True):
        """
        :default_model: model for BERTScore
        :rescale_with_baseline:
            - whether or not to rescale the score. If False, the values of BERTScore will be very high
            - this issue was observed and later added to the BERTScore package,
            - see https://github.com/Tiiiger/bert_score/blob/master/journal/rescale_baseline.md
        """
        self.nlp = spacy.load("en_core_web_sm")
        self.default_model = default_model # en => roberta-large
        self.rescale_with_baseline = rescale_with_baseline
        print("SelfCheck-BERTScore initialized")

    @torch.no_grad()
    def predict(
        self,
        sentences: List[str],
        sampled_passages: List[str],
    ):
        """
        This function takes sentences (to be evaluated) with sampled passages (evidence), and return sent-level scores
        :param sentences: list[str] -- sentences to be evaluated, e.g. GPT text response spilt by spacy
        :param sampled_passages: list[str] -- stochastically generated responses (without sentence splitting)
        :return sent_scores: sentence-level score which is 1.0 - bertscore
        """
        num_sentences = len(sentences)
        num_samples = len(sampled_passages)
        bertscore_array = np.zeros((num_sentences, num_samples))
        for s in range(num_samples):
            sample_passage = sampled_passages[s]
            sentences_sample = [sent for sent in self.nlp(sample_passage).sents] # List[spacy.tokens.span.Span]
            sentences_sample = [sent.text.strip() for sent in sentences_sample if len(sent) > 3]
            num_sentences_sample  = len(sentences_sample)

            refs  = expand_list1(sentences, num_sentences_sample) # r1,r1,r1,....
            cands = expand_list2(sentences_sample, num_sentences) # s1,s2,s3,...

            P, R, F1 = bert_score.score(
                    cands, refs,
                    lang=self.default_model, verbose=False,
                    rescale_with_baseline=self.rescale_with_baseline,
            )
            F1_arr = F1.reshape(num_sentences, num_sentences_sample)
            F1_arr_max_axis1 = F1_arr.max(axis=1).values
            F1_arr_max_axis1 = F1_arr_max_axis1.numpy()

            bertscore_array[:,s] = F1_arr_max_axis1

        bertscore_mean_per_sent = bertscore_array.mean(axis=-1)
        one_minus_bertscore_mean_per_sent = 1.0 - bertscore_mean_per_sent
        return one_minus_bertscore_mean_per_sent

class SelfCheckNgram:
    """
    SelfCheckGPT (Ngram variant): Checking LLM's text against its own sampled texts via ngram model
    Note that this variant of SelfCheck score is not bounded in [0.0, 1.0]
    """
    def __init__(self, n: int, lowercase: bool = True):
        """
        :param n: n-gram model, n=1 is Unigram, n=2 is Bigram, etc.
        :param lowercase: whether or not to lowercase when counting n-grams
        """
        self.n = n
        self.lowercase = lowercase
        print(f"SelfCheck-{n}gram initialized")

    def predict(
        self,
        sentences: List[str],
        passage: str,
        sampled_passages: List[str],
    ):
        if self.n == 1:
            ngram_model = UnigramModel(lowercase=self.lowercase)
        elif self.n > 1:
            ngram_model = NgramModel(n=self.n, lowercase=self.lowercase)
        else:
            raise ValueError("n must be integer >= 1")
        ngram_model.add(passage)
        for sampled_passge in sampled_passages:
            ngram_model.add(sampled_passge)
        ngram_model.train(k=0)
        ngram_pred = ngram_model.evaluate(sentences)
        return ngram_pred

class SelfCheckNLI:
    """
    SelfCheckGPT (NLI variant): Checking LLM's text against its own sampled texts via DeBERTa-v3 finetuned to Multi-NLI
    """
    def __init__(
        self,
        nli_model: str = None,
        device = None
    ):
        nli_model = nli_model if nli_model is not None else NLIConfig.nli_model
        self.tokenizer = DebertaV2Tokenizer.from_pretrained(nli_model)
        self.model = DebertaV2ForSequenceClassification.from_pretrained(nli_model)
        self.model.eval()
        if device is None:
            device = torch.device("cpu")
        self.model.to(device)
        self.device = device
        print("SelfCheck-NLI initialized to device", device)

    @torch.no_grad()
    def predict(
        self,
        sentences: List[str],
        sampled_passages: List[str],
    ):
        """
        This function takes sentences (to be evaluated) with sampled passages (evidence), and return sent-level scores
        :param sentences: list[str] -- sentences to be evaluated, e.g. GPT text response spilt by spacy
        :param sampled_passages: list[str] -- stochastically generated responses (without sentence splitting)
        :return sent_scores: sentence-level score which is P(condict|sentence, sample)
        note that we normalize the probability on "entailment" or "contradiction" classes only
        and the score is the probability of the "contradiction" class
        """
        num_sentences = len(sentences)
        num_samples = len(sampled_passages)
        scores = np.zeros((num_sentences, num_samples))
        for sent_i, sentence in enumerate(sentences):
            for sample_i, sample in enumerate(sampled_passages):
                inputs = self.tokenizer.batch_encode_plus(
                    batch_text_or_text_pairs=[(sentence, sample)],
                    add_special_tokens=True, padding="longest",
                    truncation=True, return_tensors="pt",
                    return_token_type_ids=True, return_attention_mask=True,
                )
                inputs = inputs.to(self.device)
                logits = self.model(**inputs).logits # neutral is already removed
                probs = torch.softmax(logits, dim=-1)
                prob_ = probs[0][1].item() # prob(contradiction)
                scores[sent_i, sample_i] = prob_
        scores_per_sentence = scores.mean(axis=-1)
        return scores_per_sentence



# adapted from https://github.com/seismatica/ngram

import spacy
import numpy as np
from nltk.util import ngrams
from typing import Dict, List, Set, Tuple, Union

class UnigramModel:
    def __init__(self, lowercase: bool = True) -> None:
        self.nlp = spacy.load("en_core_web_sm")
        self.sentence_count = 0
        self.token_count = 0
        self.counts = {'<unk>': 0}
        self.lowercase = lowercase

    def add(self, text: str) -> None:
        """
        Add/Count number of unigrams in text, one sentence at a time
        """
        sentences = [sent.text.strip() for sent in self.nlp(text).sents]
        for sentence in sentences:
            tokens = [token.text for token in self.nlp(sentence)]
            if self.lowercase:
                tokens = [token.lower() for token in tokens]
            self.sentence_count += 1
            self.token_count += len(tokens)
            for unigram in tokens:
                if unigram not in self.counts:
                    self.counts[unigram] = 1
                else:
                    self.counts[unigram] += 1

    def train(self, k: int = 0) -> None:
        """
        For each unigram in the vocab, calculate its probability in the text
        :param k: smoothing pseudo-count for each unigram
        """
        self.probs = {}
        for unigram, unigram_count in self.counts.items():
            prob_nom = unigram_count + k
            prob_denom = self.token_count + k * len(self.counts) # len(self.counts) = vocab_size
            self.probs[unigram] = prob_nom / prob_denom

    def evaluate(self, sentences: List[str]) -> float:
        """
        Calculate the negative log likelihood of the model on the evaluation sentences
        """
        avg_neg_logprob = []
        max_neg_logprob = []
        logprob_doc = [] # for computing Average at document-level, i.e. Avg(Tokens)
        for sentence in sentences:
            logprob_sent = []
            tokens = [token.text for token in self.nlp(sentence)]
            for token in tokens:
                token_ = token
                if self.lowercase:
                    token = token.lower()
                if token not in self.counts:
                    token_ = '<unk>'
                    token = '<unk>'
                train_prob = self.probs[token]
                logprob = np.log(train_prob)
                logprob_sent.append(logprob)
                logprob_doc.append(logprob)
            avg_neg_logprob += [-1.0 * np.mean(logprob_sent)]
            max_neg_logprob += [-1.0 * np.min(logprob_sent)]
        avg_neg_logprob_doc = -1.0 * np.mean(logprob_doc)
        avg_max_neg_logprob_doc = np.mean(max_neg_logprob)
        return {
            'sent_level': {'avg_neg_logprob': avg_neg_logprob, 'max_neg_logprob': max_neg_logprob},
            'doc_level': {'avg_neg_logprob': avg_neg_logprob_doc, 'avg_max_neg_logprob': avg_max_neg_logprob_doc},
        }

class NgramModel:
    def __init__(
        self, n: int, lowercase: bool = True, left_pad_symbol: str = '<s>') -> None:
        self.nlp = spacy.load("en_core_web_sm")
        self.sentence_count = 0
        self.ngram_count = 0
        self.counts = {'<unk>': 0}
        self.n = n
        self.lowercase = lowercase
        self.left_pad_symbol = left_pad_symbol

    def add(self, text: str) -> None:
        """
        Add/Count number of ngrams in text, one sentence at a time
        """
        sentences = [sent.text.strip() for sent in self.nlp(text).sents]
        for sentence in sentences:
            tokens = [token.text for token in self.nlp(sentence)]
            if self.lowercase:
                tokens = [token.lower() for token in tokens]
            ngs = list(ngrams(tokens, n=self.n, pad_left=True, left_pad_symbol=self.left_pad_symbol))
            assert len(ngs) == len(tokens)
            self.sentence_count += 1
            self.ngram_count += len(ngs)
            for ng in ngs:
                if ng not in self.counts:
                    self.counts[ng] = 1
                else:
                    self.counts[ng] += 1

    def train(self, k: int = 0) -> None:
        """
        For each ngram in the vocab, calculate its probability in the text
        :param k: smoothing pseudo-count for each ngram
        """
        self.probs = {}
        for ngram, ngram_count in self.counts.items():
            prob_nom = ngram_count + k
            prob_denom = self.ngram_count + k * len(self.counts) # len(self.counts) = vocab_size
            self.probs[ngram] = prob_nom / prob_denom

    def evaluate(self, sentences: List[str]) -> float:
        """
        Calculate the negative log likelihood of the model on the evaluation sentences
        """
        avg_neg_logprob = []
        max_neg_logprob = []
        logprob_doc = [] # for computing Average at document-level, i.e. Avg(Tokens)
        for sentence in sentences:
            logprob_sent = []
            tokens = [token.text for token in self.nlp(sentence)]
            if self.lowercase:
                tokens_ = [tok.lower() for tok in tokens]
            else:
                tokens_ = [tok for tok in tokens]
            ngs = list(ngrams(tokens_, n=self.n, pad_left=True, left_pad_symbol=self.left_pad_symbol))
            assert len(ngs) == len(tokens)
            for token, ng in zip(tokens, ngs):
                if ng not in self.counts:
                    ng = '<unk>'
                train_prob = self.probs[ng]
                logprob = np.log(train_prob)
                # item = {
                #     'token_id': token_count,
                #     'token': token, # stored the original token
                #     'logprob': logprob,
                # }
                logprob_sent.append(logprob)
                logprob_doc.append(logprob)
            avg_neg_logprob += [-1.0 * np.mean(logprob_sent)]
            max_neg_logprob += [-1.0 * np.min(logprob_sent)]
        avg_neg_logprob_doc = -1.0 * np.mean(logprob_doc)
        avg_max_neg_logprob_doc = np.mean(max_neg_logprob)
        return {
            'sent_level': {'avg_neg_logprob': avg_neg_logprob, 'max_neg_logprob': max_neg_logprob},
            'doc_level': {'avg_neg_logprob': avg_neg_logprob_doc, 'avg_max_neg_logprob': avg_max_neg_logprob_doc},
        }



import re
from typing import Dict, List, Set, Tuple, Union, Any
import numpy as np
import torch
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
from transformers import LongformerTokenizer, LongformerForMultipleChoice
from selfcheckgpt.utils import prepare_qa_input, prepare_distractor_input, prepare_answering_input
from selfcheckgpt.utils import MQAGConfig, get_prob_distances

# ---------------------------------------------------------------------------------------- #
# Functions for Question Generation & Answering
def question_generation_sentence_level(
    g1_model,
    g1_tokenizer,
    g2_model,
    g2_tokenizer,
    sentence,
    passage,
    num_questions_per_sent,
    device,
):
    qa_input_ids = prepare_qa_input(
            g1_tokenizer,
            context=sentence,
            device=device,
    )
    num_valid_questions = 0
    questions = []
    for q_ in range(num_questions_per_sent):
        # Stage G.1: question+answer generation
        outputs = g1_model.generate(
            qa_input_ids,
            max_new_tokens=128,
            do_sample=True,
        )
        question_answer = g1_tokenizer.decode(outputs[0], skip_special_tokens=False)
        question_answer = question_answer.replace(g1_tokenizer.pad_token, "").replace(g1_tokenizer.eos_token, "")
        question_answer_split = question_answer.split(g1_tokenizer.sep_token)
        if len(question_answer_split) == 2:
            # valid Question + Annswer output
            num_valid_questions += 1
        else:
            continue
        question = question_answer_split[0].strip()
        answer = question_answer_split[1].strip()

        # Stage G.2: Distractor Generation
        distractor_input_ids = prepare_distractor_input(
            g2_tokenizer,
            context = passage,
            question = question,
            answer = answer,
            device = device,
            separator = g2_tokenizer.sep_token,
        )
        outputs = g2_model.generate(
            distractor_input_ids,
            max_new_tokens=128,
            do_sample=True,
        )
        distractors = g2_tokenizer.decode(outputs[0], skip_special_tokens=False)
        distractors = distractors.replace(g2_tokenizer.pad_token, "").replace(g2_tokenizer.eos_token, "")
        distractors = re.sub("<extra\S+>", g2_tokenizer.sep_token, distractors)
        distractors = [y.strip() for y in distractors.split(g2_tokenizer.sep_token)]
        options = [answer] + distractors

        while len(options) < 4:
            # print("Warning: options =", options)
            options.append(options[-1])

        question_item = {
            'question': question,
            'options': options,
        }
        questions.append(question_item)
    return questions


def question_generation_sampling(
    g1_model,
    g1_tokenizer,
    g2_model,
    g2_tokenizer,
    context,
    num_questions,
    device,
):
    qa_input_ids = prepare_qa_input(
            g1_tokenizer,
            context=context,
            device=device,
    )
    max_repeated_sampling = int(num_questions * 1.5) # sometimes generated question+answer is invalid
    num_valid_questions = 0
    questions = []
    for q_ in range(max_repeated_sampling):
        # Stage G.1: question+answer generation
        outputs = g1_model.generate(
            qa_input_ids,
            max_new_tokens=128,
            do_sample=True,
        )
        question_answer = g1_tokenizer.decode(outputs[0], skip_special_tokens=False)
        question_answer = question_answer.replace(g1_tokenizer.pad_token, "").replace(g1_tokenizer.eos_token, "")
        question_answer_split = question_answer.split(g1_tokenizer.sep_token)
        if len(question_answer_split) == 2:
            # valid Question + Annswer output
            num_valid_questions += 1
        else:
            continue
        question = question_answer_split[0].strip()
        answer = question_answer_split[1].strip()

        # Stage G.2: Distractor Generation
        distractor_input_ids = prepare_distractor_input(
            g2_tokenizer,
            context = context,
            question = question,
            answer = answer,
            device = device,
            separator = g2_tokenizer.sep_token,
        )
        outputs = g2_model.generate(
            distractor_input_ids,
            max_new_tokens=128,
            do_sample=True,
        )
        distractors = g2_tokenizer.decode(outputs[0], skip_special_tokens=False)
        distractors = distractors.replace(g2_tokenizer.pad_token, "").replace(g2_tokenizer.eos_token, "")
        distractors = re.sub("<extra\S+>", g2_tokenizer.sep_token, distractors)
        distractors = [y.strip() for y in distractors.split(g2_tokenizer.sep_token)]
        options = [answer] + distractors

        while len(options) < 4:
            options.append(options[-1])

        question_item = {
            'question': question,
            'options': options,
        }
        questions.append(question_item)
        if num_valid_questions == num_questions:
            break
    return questions

def question_generation_beamsearch(
    g1_model,
    g1_tokenizer,
    g2_model,
    g2_tokenizer,
    context,
    num_beams,
    device,
):
    qa_input_ids = prepare_qa_input(
        g1_tokenizer,
        context=context,
        device=device,
    )
    # Stage G.1: question+answer generation
    outputs = g1_model.generate(
        qa_input_ids,
        max_new_tokens=128,
        do_sample=False,
        num_beams=num_beams,
    )
    question_answer = g1_tokenizer.decode(outputs[0], skip_special_tokens=False)
    question_answer = question_answer.replace(g1_tokenizer.pad_token, "").replace(g1_tokenizer.eos_token, "")
    question_answer_split = question_answer.split(g1_tokenizer.sep_token)
    if len(question_answer_split) == 2:
        question = question_answer_split[0].strip()
        answer = question_answer_split[1].strip()
    else:
        question = question_answer_split[0].strip()
        answer = 'none'

    # Stage G.2: Distractor Generation
    distractor_input_ids = prepare_distractor_input(
        g2_tokenizer,
        context = context,
        question = question,
        answer = answer,
        device = device,
        separator = g2_tokenizer.sep_token,
    )
    outputs = g2_model.generate(
        distractor_input_ids,
        max_new_tokens=128,
        do_sample=False,
        num_beams=num_beams,
    )
    distractors = g2_tokenizer.decode(outputs[0], skip_special_tokens=False)
    distractors = distractors.replace(g2_tokenizer.pad_token, "").replace(g2_tokenizer.eos_token, "")
    distractors = re.sub("<extra\S+>", g2_tokenizer.sep_token, distractors)
    distractors = [y.strip() for y in distractors.split(g2_tokenizer.sep_token)]
    options = [answer] + distractors

    while len(options) < 4:
        options.append(options[-1])

    question_item = {
        'question': question,
        'options': options,
    }
    return [question_item]


def answering(
    a_model,
    a_tokenizer,
    question,
    options,
    context,
    max_seq_length,
    device,
):
    answering_given_passage = prepare_answering_input(
        tokenizer=a_tokenizer,
        question=question,
        options=options,
        context=context,
        device=device,
        max_seq_length=max_seq_length,
    )
    answering_outputs = a_model(**answering_given_passage)
    prob = torch.softmax(answering_outputs['logits'], dim=-1)[0].cpu().numpy()
    return prob

# ---------------------------------------------------------------------------------------- #
# Main MQAG class
class MQAG:
    def __init__(self,
        g1_model_type: str = 'race',
        device = None,
    ):
        assert g1_model_type in ['race', 'squad']
        self.g1_model_type = g1_model_type

        if device is None:
            device = torch.device("cpu")
        self.device = device
        self.inti_generation = False
        self.inti_answering = False
        print(f"MQAG ({g1_model_type}) initialized to {device}")

    def _initialize_generation(self):
        if self.g1_model_type == 'race':
            g1_model_name = MQAGConfig.generation1_race
        elif self.g1_model_type == 'squad':
            g1_model_name = MQAGConfig.generation1_squad

        # Question Generation Systems (G1 & G2)
        self.g1_tokenizer = AutoTokenizer.from_pretrained(g1_model_name)
        self.g1_model = AutoModelForSeq2SeqLM.from_pretrained(g1_model_name)
        self.g2_tokenizer = AutoTokenizer.from_pretrained(MQAGConfig.generation2)
        self.g2_model = AutoModelForSeq2SeqLM.from_pretrained(MQAGConfig.generation2)
        self.g1_model.eval()
        self.g2_model.eval()
        self.g1_model.to(self.device)
        self.g2_model.to(self.device)
        print(f"Initialized Generation")

    def _initialize_answering(self):
        # Question Answering System (A)
        self.a_tokenizer = LongformerTokenizer.from_pretrained(MQAGConfig.answering)
        self.a_model = LongformerForMultipleChoice.from_pretrained(MQAGConfig.answering)
        self.a_model.eval()
        self.a_model.to(self.device)
        print(f"Initialized Answering")

    def score(
        self,
        candidate: str,
        reference: str,
        num_questions: int = 10,
        verbose: bool = False,
    ):
        """
        MQAG score
        :param candidate: text from which questions will be derived, e.g. the summary
        :param reference: text to be used as the ground-truth, e.g. the original document
        :return distances: dict{'kl_div': float, 'counting': float, 'hellinger': float, 'total_variation': float}
        """
        questions = self.generate(context=candidate, do_sample=True, num_questions=num_questions)
        probs_cad = self.answer(questions=questions, context=candidate)
        probs_ref = self.answer(questions=questions, context=reference)
        kl_, ct_, hl_, tv_ = 0, 0, 0, 0
        for i in range(num_questions):
            p1 = probs_cad[i]
            p2 = probs_ref[i]
            kl, ct, hl, tv = get_prob_distances(p1, p2)
            kl_ += kl
            ct_ += ct
            hl_ += hl
            tv_ += tv
        kl_ = kl_ / num_questions
        ct_ = ct_ / num_questions
        hl_ = hl_ / num_questions
        tv_ = tv_ / num_questions
        distances = {'kl_div': kl_, 'counting': ct_, 'hellinger': hl_, 'total_variation': tv_}
        if verbose:
            for i in range(num_questions):
                question, options = questions[i]['question'], questions[i]['options']
                print(f"Q{i+1}: {question}")
                print("(1) [P(.|cand)={:.2f}%]\t[P(.|ref)={:.2f}%]\t{}".format(probs_cad[i][0]*100,  probs_ref[i][0]*100, options[0]))
                print("(2) [P(.|cand)={:.2f}%]\t[P(.|ref)={:.2f}%]\t{}".format(probs_cad[i][1]*100,  probs_ref[i][1]*100, options[1]))
                print("(3) [P(.|cand)={:.2f}%]\t[P(.|ref)={:.2f}%]\t{}".format(probs_cad[i][2]*100,  probs_ref[i][2]*100, options[2]))
                print("(4) [P(.|cand)={:.2f}%]\t[P(.|ref)={:.2f}%]\t{}".format(probs_cad[i][3]*100,  probs_ref[i][3]*100, options[3]))
                print("-------------------------------------------------------------------------------")
        return distances

    @torch.no_grad()
    def generate(
        self,
        context: str,
        do_sample: bool = True,
        num_questions: int = 5,
        **kwargs
    ):
        if self.inti_generation == False:
            self._initialize_generation()
            self.inti_generation = True
        if do_sample:
            questions = question_generation_sampling(
                self.g1_model, self.g1_tokenizer,
                self.g2_model, self.g2_tokenizer,
                context, num_questions, self.device,
            )
        else: # beam_search decoding
            if num_questions != 1:
                print("warning: do_sample is False ---> only 1 sample will be generated")
            if 'num_beams' in kwargs:
                num_beams = kwargs['num_beams']
            else:
                num_beams = 5
            questions = question_generation_beamsearch(
                self.g1_model, self.g1_tokenizer,
                self.g2_model, self.g2_tokenizer,
                context, num_beams, self.device,
            )
        return questions

    @torch.no_grad()
    def answer(
        self,
        questions: List[Dict[str, Any]],
        context: str,
    ):
        """
        :param quetions: List of x where x = {'question': str, 'options': List[str]}
        :param context: string
        :return probs: np.array of dimension (num_questions, 4)
        """
        if self.inti_answering == False:
            self._initialize_answering()
            self.inti_answering = True

        num_questions = len(questions)
        probs = np.zeros((num_questions, 4))
        for i, question_item in enumerate(questions):
            question, options = question_item['question'], question_item['options']
            prob = answering(
                self.a_model, self.a_tokenizer,
                question, options, context,
                max_seq_length=4096, device=self.device,
            )
            probs[i] = prob
        return probs



from setuptools import setup

# reading long description from file
with open('DESCRIPTION.txt') as file:
    long_description = file.read()

REQUIREMENTS = [
    "transformers>=4.11.3",
    "torch>=1.10",
    "numpy",
    "bert_score",
    "spacy",
    "nltk",
]

# some more details
CLASSIFIERS = [
    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 3",
]

VERSION = {}
with open("selfcheckgpt/version.py", "r") as version_file:
    exec(version_file.read(), VERSION)

# calling the setup function
setup(
    name='selfcheckgpt',
    version=VERSION["__version__"],
    description='SelfCheckGPT: Assessing text-based responses from LLMs',
    long_description=long_description,
    url='https://github.com/potsawee/selfcheckgpt',
    author='Potsawee Manakul',
    author_email='m.potsawee@gmail.com',
    license='MIT',
    packages=['selfcheckgpt'],
    classifiers=CLASSIFIERS,
    install_requires=REQUIREMENTS,
    keywords='selfcheckgpt',
    include_package_data=True,
)



import numpy as np

class MQAGConfig:
    generation1_squad: str = "potsawee/t5-large-generation-squad-QuestionAnswer"
    generation1_race: str = "potsawee/t5-large-generation-race-QuestionAnswer"
    generation2: str = "potsawee/t5-large-generation-race-Distractor"
    answering: str = "potsawee/longformer-large-4096-answering-race"
    answerability: str = "potsawee/longformer-large-4096-answerable-squad2"

class NLIConfig:
    nli_model: str = "potsawee/deberta-v3-large-mnli"

# Question Generation & Answering Input Processing
def prepare_qa_input(t5_tokenizer, context, device):
    """
    input: context
    output: question <sep> answer
    """
    encoding = t5_tokenizer(
        [context],
        return_tensors="pt",
    )
    input_ids = encoding.input_ids.to(device)
    return input_ids


def prepare_distractor_input(t5_tokenizer, context, question, answer, device, separator='<sep>'):
    """
    input: question <sep> answer <sep> article
    output: distractor1 <sep> distractor2 <sep> distractor3
    """
    input_text = question + ' ' + separator + ' ' + answer + ' ' + separator + ' ' + context
    encoding = t5_tokenizer(
        [input_text],
        return_tensors="pt",
    )
    input_ids = encoding.input_ids.to(device)
    return input_ids


def prepare_answering_input(
    tokenizer, # longformer_tokenizer
    question, options, context,
    device, max_seq_length=4096,
):
    c_plus_q = context + ' ' + tokenizer.bos_token + ' ' + question
    c_plus_q_4 = [c_plus_q] * len(options)

    tokenized_examples = tokenizer(
        c_plus_q_4, options,
        max_length=max_seq_length,
        padding="longest",
        truncation=True,
        return_tensors="pt",
    )
    tokenized_examples = tokenized_examples.to(device)
    input_ids = tokenized_examples['input_ids'].unsqueeze(0)
    attention_mask = tokenized_examples['attention_mask'].unsqueeze(0)

    example_encoded = {
        "input_ids": input_ids,
        "attention_mask": attention_mask,
    }

    return example_encoded

# SelfCheck - BERTScore utils
def expand_list1(mylist, num):
    expanded = []
    for x in mylist:
        for _ in range(num):
            expanded.append(x)
    return expanded

def expand_list2(mylist, num):
    expanded = []
    for _ in range(num):
        for x in mylist:
            expanded.append(x)
    return expanded

# MQAG score utils
def smoothing(probs):
    probs = probs + 1e-12
    probs = probs / probs.sum()
    return probs

def kl_div(probs1, probs2):
    assert len(probs1) == len(probs2)
    probs1 = smoothing(probs1)
    probs2 = smoothing(probs2)
    xx = probs1 * np.log(probs1 / probs2)
    return xx.sum()

def onebest_argmax(probs1, probs2):
    answer1 = probs1.argmax()
    answer2 = probs2.argmax()
    if answer1 == answer2:
        count = 0
    else:
        count = 1
    return count

def hellinger_dist(probs1, probs2):
    # https://en.wikipedia.org/wiki/Hellinger_distance
    sqrt_p1 = np.sqrt(probs1)
    sqrt_p2 = np.sqrt(probs2)
    return ((sqrt_p1 - sqrt_p2)**2).sum(axis=-1) / 1.4142135

def total_variation(probs1, probs2):
    diff = np.abs(probs1 - probs2)
    return diff.max()

def get_prob_distances(probs1, probs2):
    kl = kl_div(probs1, probs2)
    ob = onebest_argmax(probs1, probs2)
    hl = hellinger_dist(probs1, probs2)
    tv = total_variation(probs1, probs2)
    return kl, ob, hl, tv
